#include "Vm.h"
#include <iostream>

namespace JV
{
    VM::VM(const std::vector<uint8_t> &bytecode)
        : m_Bytecode(bytecode), m_IP(0)
    {
        m_Registers.fill(0);
        m_Heap.resize(HEAP_SIZE, 0);
    }

    int VM::Pop()
    {
        int val = m_Stack.top();
        m_Stack.pop();
        return val;
    }

    void VM::Push(int val)
    {
        m_Stack.push(val);
    }

    void VM::Debug()
    {
        std::cout << "\n[DEBUG] IP: " << m_IP << "\nStack: ";
        std::stack<int> temp = m_Stack;
        std::vector<int> stackDump;
        while (!temp.empty()) {
            stackDump.push_back(temp.top());
            temp.pop();
        }
        for (auto it = stackDump.rbegin(); it != stackDump.rend(); ++it)
            std::cout << *it << " ";
        std::cout << "\nRegisters: ";
        for (int r: m_Registers)
            std::cout << r << " ";
        std::cout << "\n\nPress Enter to step...\n";
        std::cin.get();
    }

    void VM::Run()
    {
        while (m_IP < m_Bytecode.size()) {
            OpCode op = static_cast<OpCode>(m_Bytecode[m_IP++]);
#ifndef _DEBUG
            //Debug();
#endif

            switch (op) {
                case OP_HALT:
                    return;

                case OP_ALLOC: Alloc();
                    break;

                case OP_STORE_MEM: StoreMem();
                    break;
                case OP_LOAD_MEM: LoadMem();
                    break;


                case OP_CONST:
                    Push(m_Bytecode[m_IP++]);
                    break;

                case OP_STORE:
                {
                    uint8_t reg = m_Bytecode[m_IP++];
                    m_Registers[reg] = Pop();
                    break;
                }

                case OP_LOAD:
                {
                    uint8_t reg = m_Bytecode[m_IP++];
                    Push(m_Registers[reg]);
                    break;
                }

                case OP_ADD:
                {
                    int b = Pop(), a = Pop();
                    Push(a + b);
                    break;
                }

                case OP_SUB:
                {
                    int b = Pop(), a = Pop();
                    Push(a - b);
                    break;
                }

                case OP_MUL:
                {
                    int b = Pop(), a = Pop();
                    Push(a * b);
                    break;
                }

                case OP_DIV:
                {
                    int b = Pop(), a = Pop();
                    Push(a / b);
                    break;
                }

                case OP_PRINT:
                    std::cout << Pop() << std::endl;
                    break;

                case OP_CALL:
                {
                    uint8_t addr = m_Bytecode[m_IP++];
                    m_CallStack.push(m_IP);
                    m_IP = addr;
                    break;
                }

                case OP_RET:
                    m_IP = m_CallStack.top();
                    m_CallStack.pop();
                    break;

                case OP_DEBUG:
                    Debug();
                    break;

                default:
                    std::cerr << "Unknown opcode: " << static_cast<int>(op) << "\n";
                    return;
            }
        }
    }


    void VM::Alloc()
    {
        uint8_t count = m_Bytecode[m_IP++];
        if (m_HeapPtr + count >= HEAP_SIZE) {
            std::cerr << "Heap overflow!" << std::endl;
            return;
        }
        Push(static_cast<int>(m_HeapPtr)); // return address
        m_HeapPtr += count;
    }

    void VM::StoreMem()
    {
        int value = Pop();
        int addr = Pop();
        if (addr >= 0 && addr < HEAP_SIZE)
            m_Heap[addr] = value;
        else
            std::cerr << "Invalid memory store at " << addr << std::endl;
    }

    void VM::LoadMem()
    {
        int addr = Pop();
        if (addr >= 0 && addr < HEAP_SIZE)
            Push(m_Heap[addr]);
        else
            std::cerr << "Invalid memory load at " << addr << std::endl;
    }
} // namespace JV
